/*
 * 29/03/2004
 *
 * image.cpp
 * Thierry Vouriot (Yri)
 * yeri@fr.st
 * http://www.yeri.fr.st
 * thierry.vouriot@free.fr
 * http://thierry.vouriot.free.fr
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "image.h"

// ============
// Constructeur
// ============
CImage::CImage() {
	m_data = NULL;
	m_width = 0;
	m_height = 0;
	m_bitCount = 0;
}

// ============
// Constructeur
// ============
CImage::CImage(int w, int h, int d)
{
	int nbplan = d / 8;
	m_data = new unsigned char[w * h * nbplan];
	m_width = w;
	m_height = h;
	m_bitCount = d;
}

// ============
// Constructeur
// ============
CImage::CImage(unsigned char *da, int w, int h, int d)
{
	int nbplan = d / 8;
	m_data = new unsigned char[w * h * nbplan];
	memcpy((void*)m_data, (void*)da, w * h * nbplan * sizeof(unsigned char));
	m_width = w;
	m_height = h;
	m_bitCount = d;
}

// ===========
// Destructeur
// ===========
CImage::~CImage()
{
	if (m_data) delete m_data;
}

// =============================
// Retourne une copie de l'image
// =============================
CImage * CImage::Copy()
{
	return new CImage(m_data, m_width, m_height, m_bitCount);
}
	

// ======================================================
// Fonction lisant un fichier BMP 8 ou 24 (non compresse)
// ======================================================
int CImage::Load(char *filename)
{
	int type = GetType(filename);
	switch (type) {
		case BMP:
			return LoadBMP(this, filename);	
		case TGA:
			return LoadTGA(this, filename);
		default:
			return IMG_BAD_IMAGE_TYPE;
	}

}

// ==================================================
// Retourne le type de l'image en fonction de son nom
// ==================================================
int CImage::GetType(const char *name)
{
	int i;
	int res;
	for (i = 0 ; name[i] != '\0'; i++);

	while (name[i] != '.' || i==0)
	{
		i--;
	}
	
	if (i==0) return INVALID;

	char *ext;
	
#ifdef WIN32
	ext = strupr(strdup(name+i));
	if (strcmp(ext, ".BMP") == 0) res = BMP;
	else if (strcmp(ext, ".TGA") == 0) res = TGA;
	else res = INVALID;
#else
	ext = strdup(name+i);
	if (strcasecmp(ext, ".BMP") == 0) res = BMP;
	else if (strcasecmp(ext, ".TGA") == 0) res = TGA;
	else res = INVALID;
#endif

	delete ext;

	return res;
}

#define CTOI(C) (*(int*)&C)	//rcupre en int un nombre point par un char*

// ======================================================
// Fonction lisant un fichier BMP 8 ou 24 (non compresse)
// ======================================================
int CImage::LoadBMP(CImage *img, const char *filename) 
{
	FILE *filePtr;			// pointeur sur le fichier
	unsigned char header[0x36];	// tableau qui contiendra l'en-tte du fichier BMP
	int colorMode;			// 3 pour RGB
	int imagePos;			// offset du dbut des donnes de l'image
	long imageSize;			// taille de l'image BMP
	long imageIdx;			// compteur
	unsigned char colorSwap;	// variable d'change
	
	// ouverture du fichier BMP
	if (!(filePtr = fopen(filename, "rb"))) {
        	return IMG_FILE_NOT_FOUND;
	}	
	// lecture de l'en-tte
	if (fread(header,1,0x36,filePtr)!=0x36) {
		fclose(filePtr);
		return IMG_BAD_IMAGE_TYPE;
	}	
	// BM indique que l'on se trouve en prsence d'un fichier bitmap
	if (header[0]!='B' || header[1]!='M') {
		fclose(filePtr);
		return IMG_BAD_IMAGE_TYPE;
	}
	// indique que l'image n'est pas compresse	
	if (CTOI(header[0x1E])!=0) {
		fclose(filePtr);
		return IMG_BAD_IMAGE_TYPE;
	}
	// image 24 et 8 bits uniquement
	img->m_bitCount = CTOI(header[0x1C]);
	if (img->m_bitCount!=24 && img->m_bitCount!=8) {
		fclose(filePtr);
		return IMG_BAD_BITS;
	}
	colorMode = img->m_bitCount / 8;

	// rcupre les infos sur les donnes de l'image
	imagePos = CTOI(header[0x0A]);
	imageSize = CTOI(header[0x22]);
	// rcupre les dimensions de l'image
	img->m_width = CTOI(header[0x12]);
	img->m_height = CTOI(header[0x16]);	
		
	// si les infos sur les donnes de l'image n'ont pas t rcupr
	// on les calcule
	if (imageSize==0)
		imageSize=img->m_height*img->m_width*colorMode;
	if (imagePos==0)
		imagePos=0x36;

	// charge les donnes de l'image
	fseek(filePtr,imagePos,0);
	img->m_data = (unsigned char*) malloc(sizeof(unsigned char) * imageSize);
	
	if (!img->m_data) {
		fclose(filePtr);
		return IMG_BAD_DATA;
	}

	if (fread(img->m_data,1,imageSize,filePtr) != (unsigned long)imageSize) 
	{
		free(img->m_data);
		fclose(filePtr);
		return IMG_BAD_DATA;
	}
	
	fclose(filePtr);

	// dans les images BMP les couleurs des pixels sont en mode BGR
	// transformation en mode RGB
	if (img->m_bitCount == 24)
	for (imageIdx=0; imageIdx < imageSize; imageIdx += colorMode)
	{
		colorSwap = img->m_data[imageIdx];
		img->m_data[imageIdx] = img->m_data[imageIdx + 2];
		img->m_data[imageIdx + 2] = colorSwap;
	}
	
	// Format de l'image RGB ou niveau de gris
	switch (img->m_bitCount) {
		case 8: img->m_texFormat = GL_LUMINANCE; break;
		case 24: img->m_texFormat = GL_RGB; break;
	}
	
	return IMG_OK;
}

// ================================================
// Fonction permettant de lire un fichier image TGA
// 8, 16, 24 ou 32 bits
// Le format 32 bits est convertit en 24 bits
// ================================================
// Le format Targa TGA est divis en 2 partie l'en-tte et les donnes.
// L'en-tte contient 12 champs et est dfinie par la structure suivante :
//
// typedef struct tagTARGAFILEHEADER
// {
//	unsigned char imageIDLength;	// number of characters in identification field;
//									// 0 denotes no identification field is included
//	unsigned char colorMapType;		// type of color map; always 0
//	unsigned char imageTypeCode;	// uncompressed RGB is 2;
//									// uncompressed grayscale is 3
//	short int colorMapOrigin;		// origin of color map (lo-hi); always 0
//	short int colorMapLength;		// length of color map (lo-hi); always 0
//	short int colorMapEntrySize;	// color map entry size (lo-hi); always 0
//	short int imageXOrigin;			// x coordinate of lower-left corner of image
//									// (lo-hi); always 0
//	short int imageYOrigin;			// y coordinate of lower-left corner of image
//									// (lo-hi); always 0
//	short int imageWidth;			// width of image in pixels (lo-hi)
//	short int imageHeight;			// height of image in pixels (lo-hi)
//	unsigned char bitCount;			// number of bits; 16, 24, 32
//	unsigned char imageDescriptor;	// 24 bit = 0x00; 32 bit = 0x08
// } TARGAFILEHEADER
int CImage::LoadTGA(CImage *img, const char *filename)
{
	FILE *filePtr;				// pointeur sur le fichier
	unsigned char ucharBad;		// poubelle pour les unsigned char
	short int sintBad;			// poubelle pour les short int
	int colorMode;				// 4 pour RGBA ou 3 pour RGB
	long imageSize;				// taille de l'image TGA
	long imageIdx;				// compteur
	unsigned char colorSwap;	// variable d'change
	unsigned short pixels = 0;	// un pixel d'une image 16 bits
	int r, g, b;				// composantes RGB
	
	// ouverture du fichier TGA
	if (!(filePtr = fopen (filename, "rb"))) {
        return IMG_FILE_NOT_FOUND;
	}
	
	// lecture des 2 premiers champs dont on n'a pas besoin
	fread(&ucharBad, sizeof(unsigned char), 1, filePtr);
	fread(&ucharBad, sizeof(unsigned char), 1, filePtr);
	
	// lecture du type de l'image
	fread(&ucharBad, sizeof(unsigned char), 1, filePtr);
	if ((ucharBad != 2) && (ucharBad != 3)) {
		fclose(filePtr);
		return IMG_BAD_IMAGE_TYPE;
	}
	// lecture des 5 champs suivants dont on n'a pas besoin
	fread(&sintBad, sizeof(short int), 1, filePtr);
	fread(&sintBad, sizeof(short int), 1, filePtr);
	fread(&ucharBad, sizeof(unsigned char), 1, filePtr);
	fread(&sintBad, sizeof(short int), 1, filePtr);
	fread(&sintBad, sizeof(short int), 1, filePtr);
	
	// lecture des dimensions de l'image
	fread(&(img->m_width), sizeof(short int), 1, filePtr);
	fread(&(img->m_height), sizeof(short int), 1, filePtr);
	fread(&(img->m_bitCount), sizeof(unsigned char), 1, filePtr);
	
	// lecture du dernier champ dont on n'a pas besoin
	fread(&ucharBad, sizeof(unsigned char), 1, filePtr);
	
	// on ne lit que les images 8, 16, 24 ou 32 bits
	if (img->m_bitCount != 32 && img->m_bitCount != 24 && img->m_bitCount != 8 && img->m_bitCount != 16) {
		fclose(filePtr);
		return IMG_BAD_BITS;
	}
	
	// image 8, 24 ou 32 bits
	if(img->m_bitCount == 24 || img->m_bitCount == 8 || img->m_bitCount == 32)
	{
		// colorMode -> 3 = BGR, 4 = BGRA
		colorMode = img->m_bitCount / 8;
		imageSize = img->m_width * img->m_height * colorMode;
	
		// allocation de la mmoire pour les donnes de l'image
		//m_data = (unsigned char*) malloc(sizeof(unsigned char) * imageSize);
		img->m_data = new unsigned char[imageSize];

		// lecture des donnes de l'image
		fread(img->m_data, sizeof(unsigned char), imageSize, filePtr);
		if (img->m_data == NULL) {
        		fclose(filePtr);
			return IMG_BAD_DATA;
		}
	
		// dans les images TGA les couleurs des pixels sont en mode BGR
		// transformation en mode RGB
		if (img->m_bitCount != 8)
		for (imageIdx=0; imageIdx < imageSize; imageIdx += colorMode)
		{
			colorSwap = img->m_data[imageIdx];
			img->m_data[imageIdx] = img->m_data[imageIdx + 2];
			img->m_data[imageIdx + 2] = colorSwap;
		}
	}
	else { // image 16 bits

		img->m_bitCount = 24;
		colorMode = 3;	// mode RGB
		imageSize = img->m_width * img->m_height * colorMode;
		
		// allocation de la mmoire pour les donnes de l'image
		//tgaFile->imageData = (unsigned char*) malloc(sizeof(unsigned char) * imageSize);
		img->m_data = new unsigned char[imageSize];
		
		// en 16 bits chaque pixel est contenu dans un "unsigned short"
		// boucle sur tous les pixels de l'image
		for (imageIdx = 0; imageIdx < img->m_width * img->m_height; imageIdx++)
		{
			// lecture du pixel courant
			fread(&pixels, sizeof(unsigned short), 1, filePtr);
				
			// pour convertir 16 bits en composantes RGB
			// chaque composante occupe 5 bits
			// ex: pixels = 0x61E8 = 0110000111101000
			// on utilise un masque 0x1F = 11111
			// 0110000111101000 & 11111 = 01000
			// on effectue un dcallage de 3 bits pour avoir un nombre sur 8 bits
			// 01000 << 3 = 01000000 = 0x40 = 64 --> composante bleu
			// on dcale ensuite de 5 bits pour arriver  la prochaine
			// composante et on ritre la mme opration
			// 0110000111101000 >> 5 = 0000001100001111
			// 0000001100001111 & 11111 = 01111
			// 01111 << 3 = 01111000 = 0x78 = 120 --> composante vert
			// ...etc...
			b = (pixels & 0x1f) << 3;
			g = ((pixels >> 5) & 0x1f) << 3;
			r = ((pixels >> 10) & 0x1f) << 3;
				
			// on assigne ensuite les composantes au tableau de donnes
			img->m_data[imageIdx * 3 + 0] = r;
			img->m_data[imageIdx * 3 + 1] = g;
			img->m_data[imageIdx * 3 + 2] = b;
		}
	}

	// Format de l'image RGB ou niveau de gris
	switch (img->m_bitCount) {
		case 8: img->m_texFormat = GL_LUMINANCE; break;
		case 24: img->m_texFormat = GL_RGB; break;
		case 32: img->m_texFormat = GL_RGBA; break;
	}

	// fermeture du fichier
	fclose(filePtr);
	
	return IMG_OK;
}

// ============================================================
// Fonction enregistrant une image TGA dans le fichier filename
// ============================================================
int CImage::SaveTGA(CImage *img, const char *filename)
{
	unsigned char byteSkip;		// utilis pour mettre  0 certains champs
	short int shortSkip;		// utilis pour mettre  0 certains champs
	unsigned char imageType;	// type de l'image
	int colorMode;				// color mode
	unsigned char colorSwap;	// utilis pour la conversion RGB - BGR
	int imageIdx;				// compteur
	unsigned char bitDepth;		// image 24 bits
	long imageSize;				// taille de l'image
	FILE *filePtr;				// pointeur sur le fichier
	
	// cration du fichier en mode criture binaire
	if (!(filePtr = fopen (filename, "wb"))) {
		return IMG_FILE_NOT_CREATE;
	}
	
	imageType = 2;	// RGB uncompressed
	bitDepth = img->m_bitCount;	// bitdepth
	colorMode = img->m_bitCount / 8;	// color mode
	
	byteSkip = 0;
	shortSkip = 0;
	
	// criture des 2 premiers champs  0
	fwrite(&byteSkip, sizeof(unsigned char), 1, filePtr);
	fwrite(&byteSkip, sizeof(unsigned char), 1, filePtr);
	
	// criture du type de l'image
	fwrite(&imageType, sizeof(unsigned char), 1, filePtr);
	
	// criture des 5 champs suivants  0
	fwrite(&shortSkip, sizeof(short int), 1, filePtr);
	fwrite(&shortSkip, sizeof(short int), 1, filePtr);
	fwrite(&byteSkip, sizeof(unsigned char), 1, filePtr);
	fwrite(&shortSkip, sizeof(short int), 1, filePtr);
	fwrite(&shortSkip, sizeof(short int), 1, filePtr);
	
	// criture des dimensions de l'image
	fwrite(&img->m_width, sizeof(short int), 1, filePtr);
	fwrite(&img->m_height, sizeof(short int), 1, filePtr);
	fwrite(&bitDepth, sizeof(unsigned char), 1, filePtr);
	
	// criture du dernier champ  0
	fwrite(&byteSkip, sizeof(unsigned char), 1, filePtr);
	
	// calcul de la taille de l'image
	imageSize = img->m_width * img->m_height * colorMode;
	
	CImage *temp = img->Copy();

	// conversion des donnes du mode RGB vers le mode BGR
	if (colorMode == 3)
	for (imageIdx = 0; imageIdx < imageSize ; imageIdx += colorMode)
	{
		colorSwap = temp->m_data[imageIdx];
		temp->m_data[imageIdx] = temp->m_data[imageIdx + 2];
		temp->m_data[imageIdx + 2] = colorSwap;
	}
	
	// Symtrie d'axe verticale
	int x,y;
	int width = img->m_width*colorMode;
	for (long i = 0; i < imageSize; i+=colorMode) {
		x = i % width;
		y = i / width;
		temp->m_data[(img->m_height-1-y)*width + x] = img->m_data[i];
		if (colorMode == 3) {
			temp->m_data[(img->m_height-1-y)*width + x +1] = img->m_data[i+1];
			temp->m_data[(img->m_height-1-y)*width + x +2] = img->m_data[i+2];
		}
	}

	// criture des donnes de l'image dans le fichier
	fwrite(temp->m_data, sizeof(unsigned char), imageSize, filePtr);
	
	delete temp;

	// fermeture du fichier
	fclose(filePtr);
	
	return IMG_OK;	
}


// ====================================================================
// Retourne le format OpenGL de l'image (GL_RGB, GL_RGBA, GL_LUMINANCE)
// ====================================================================
GLenum CImage::GetTextureFormat()
{
	return m_texFormat;
}


// ==================================================
// Retourne le nombre de pixels en largeur de l'image
// ==================================================
short int CImage::GetWidth()
{
	return m_width;
}


// ==================================================
// Retourne le nombre de pixels en hauteur de l'image
// ==================================================
short int CImage::GetHeight()
{
	return m_height;
}


// ===============================================================================
// Retourne le nombre de bits par pixel (RGB = 24, RGBA = 32, Niveaux de gris = 8)
// ===============================================================================
unsigned char CImage::GetBitCount()
{
	return m_bitCount;
}


// ===================================================
// Retourne le tableau contenant les pixels de l'image
// ===================================================
unsigned char * CImage::GetData()
{
	return m_data;
}


// ================================================================
// Fonction permettant de tester si un entier est une puisance de 2
// ================================================================
bool CImage::CheckSize(int x)
{
    if (x == 2	 || x == 4 || 
        x == 8	 || x == 16 || 
        x == 32  || x == 64 ||
        x == 128 || x == 256 || 
	x == 512 || x == 1024)
        return true;
    else return false;
}


// ===================================================
// Creation d'une texture a partir de l'image courante
// ===================================================
int CImage::CreateTexture(GLuint textureArray[], GLuint textureID, bool mipmapping, GLint textureWrap)
{
	// Generate a texture with the associative texture ID stored in the array
	glGenTextures(1, &textureArray[textureID]);

	glBindTexture(GL_TEXTURE_2D, textureArray[textureID]);
		
	if (!mipmapping) {
		// filtrage plus rapide
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		
		glTexImage2D(GL_TEXTURE_2D, 0,  m_texFormat, m_width, m_height, 
			0,  m_texFormat, GL_UNSIGNED_BYTE, m_data);
	}
	else {
		// Mipmapping: meilleur qualit d'affichage mais plus lent
		gluBuild2DMipmaps(GL_TEXTURE_2D, m_texFormat, m_width, m_height, m_texFormat, GL_UNSIGNED_BYTE, m_data);

		// filtrage offrant une meilleur qualit	
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR_MIPMAP_LINEAR);
	}
	
	// l'option GL_CLAMP_TO_EDGE permet de ne pas rpter la texture comme GL_REPEAT
	// et vite donc des effets indsirables, de plus elle permet de ne pas voir
	// les bordures des textures contrairement  GL_CLAMP
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, textureWrap);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, textureWrap);

	return 1;
}


// ================================================
// Creation d'une texture a partir d'un fichier BMP
// ================================================
int CImage::CreateTexture(char* filename, GLuint textureArray[], GLuint textureID, bool mipmapping, GLint textureWrap)
{
	CImage bmp = CImage();
	int error;
	if ((error = bmp.Load(filename)) != IMG_OK) return error;
	
	// Generate a texture with the associative texture ID stored in the array
	glGenTextures(1, &textureArray[textureID]);

	glBindTexture(GL_TEXTURE_2D, textureArray[textureID]);
		
	if (!mipmapping) {
		// filtrage plus rapide
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		
		glTexImage2D(GL_TEXTURE_2D, 0,  bmp.m_texFormat, bmp.m_width, bmp.m_height, 
			0,  bmp.m_texFormat, GL_UNSIGNED_BYTE, bmp.m_data);
	}
	else {
		// Mipmapping: meilleur qualit d'affichage mais plus lent
		gluBuild2DMipmaps(GL_TEXTURE_2D, bmp.m_texFormat, bmp.m_width, bmp.m_height, bmp.m_texFormat, GL_UNSIGNED_BYTE, bmp.m_data);

		// filtrage offrant une meilleur qualit	
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR_MIPMAP_LINEAR);
	}
	
	// l'option GL_CLAMP_TO_EDGE permet de ne pas rpter la texture comme GL_REPEAT
	// et vite donc des effets indsirables, de plus elle permet de ne pas voir
	// les bordures des textures contrairement  GL_CLAMP
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, textureWrap);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, textureWrap);

	return IMG_OK;
}


/* Structure d'un fichier BMP 24 bits
 * Le format BMP est divis en 3 parties le header,
 * la zone bitmap info et les donnes
 *
 * Le header
 * Position		Type		Description
 * 00h			DB "BM"		BM indique qu'il s'agit d'un fichier bitmap
 * 02h			DD ?		Taille du fichier en octets
 * 06h			DW 0		Rserv
 * 08h			DW 0		Rserv
 * 0Ah			DD ?		Offset des donnes de l'image
 *
 * La zone bitmap info
 * Position		Type		Description
 * OEh			DD ?		Taille de la zone bitmap info
 * 12h			DD ?		Largeur de l'image en pixels
 * 16h			DD ?		Hauteur de l'image en pixels
 * 1Ah			DW 1		Nombre de plans (1)
 * 1Ch			DW 1|4|8|24	Bits par pixel
 * 1Eh			DD 0|1|2	Format de compression (O=non compresse)
 * 22h			DD ?		Taille de l'image en octets
 * 26h			DD ?		Rsolution horizontale en pixels par mtre
 * 2Ah			DD ?		Rsolution verticale en pixels par mtre
 * 2Eh			DD ?		Nombre de couleurs utilises  ...
 * 32h			DD ?		Nombre d'index de couleurs importants pour l'affichage
 *
 */
 
 
 
 
