/*
 * 05/28/2003
 *
 * image.h
 * Classe permettant la lecture, l'ecriture d'image
 * ainsi que la creation de textures 
 *
 * 2003 Thierry Vouriot (Yri)
 * yeri@fr.st
 * http://www.yeri.fr.st
 * thierry.vouriot@free.fr
 * http://thierry.vouriot.free.fr
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _BMP_H
#define _BMP_H

#ifdef _WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glu.h>
#include <cstdio>
#include <cstdlib>
#include <cstring>

#ifndef GL_CLAMP_TO_EDGE
#define GL_CLAMP_TO_EDGE 0x812F	
#endif

// Codes erreurs de retour des fonctions
enum IMGErrors
{
	IMG_FILE_NOT_FOUND,
	IMG_BAD_IMAGE_TYPE,
	IMG_BAD_BITS,
	IMG_BAD_DATA,
	IMG_FILE_NOT_CREATE,
	IMG_BAD_SIZE,
	IMG_OK
};

// Diffrents types d'images
enum IMGTypes
{
	BMP,
	TGA,
	INVALID
};

// ###########
// Classe CBmp
// ###########
class CImage
{
public:

	// ============
	// Constructeur
	// ============
	CImage();

	// ============
	// Constructeur
	// ============
	CImage(int w, int h, int d);

	// ============
	// Constructeur
	// ============
	CImage(unsigned char *da, int w, int h, int d);

	// ===========
	// Destructeur
	// ===========
	~CImage();

	// =============================
	// Retourne une copie de l'image
	// =============================
	CImage * Copy();
	
	// ======================================================
	// Fonction lisant un fichier BMP 8 ou 24 (non compresse)
	// ======================================================
	int Load(char *filename);

	// ==================================================
	// Retourne le type de l'image en fonction de son nom
	// ==================================================
	int static GetType(const char *name);

	// ====================================================================
	// Retourne le format OpenGL de l'image (GL_RGB, GL_RGBA, GL_LUMINANCE)
	// ====================================================================
	GLenum GetTextureFormat();

	// ==================================================
	// Retourne le nombre de pixels en largeur de l'image
	// ==================================================
	short int GetWidth();

	// ==================================================
	// Retourne le nombre de pixels en hauteur de l'image
	// ==================================================
	short int GetHeight();

	// ===============================================================================
	// Retourne le nombre de bits par pixel (RGB = 24, RGBA = 32, Niveaux de gris = 8)
	// ===============================================================================
	unsigned char GetBitCount();

	// ===================================================
	// Retourne le tableau contenant les pixels de l'image
	// ===================================================
	unsigned char * GetData();

	// ================================================================
	// Fonction permettant de tester si un entier est une puisance de 2
	// ================================================================
	bool static CheckSize(int x);

	int static CImage::LoadBMP(CImage *img, const char *filename);
	int static CImage::LoadTGA(CImage *img, const char *filename);
	int static CImage::SaveTGA(CImage *img, const char *filename);

	// ===================================================
	// Creation d'une texture a partir de l'image courante
	// ===================================================
	int CreateTexture(GLuint textureArray[], GLuint textureID, bool mipmapping, GLint textureWrap = GL_REPEAT);

	// ================================================
	// Creation d'une texture a partir d'un fichier BMP
	// ================================================
	int static CreateTexture(char* filename, GLuint textureArray[], GLuint textureID, bool mipmapping, GLint textureWrap = GL_REPEAT);
	
private:

	// Format OpenGL de l'image (GL_RGB, GL_RGBA, GL_LUMINANCE)
	GLenum m_texFormat;

	// Nombre de pixels en largeur de l'image
	short int m_width;

	// Nombre de pixels en hauteur de l'image
	short int m_height;

	// Nombre de bits par pixel (RGB = 24, RGBA = 32, Niveaux de gris = 8)
	unsigned char m_bitCount;

	// Tableau contenant les pixels de l'image
	unsigned char *m_data;

};

#endif



