/*
 * 29/03/2004
 *
 * image3D.cpp
 * Thierry Vouriot (Yri)
 * yeri@fr.st
 * http://www.yeri.fr.st
 * thierry.vouriot@free.fr
 * http://thierry.vouriot.free.fr
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "image3D.h"

// ============
// Constructeur
// ============
CImage3D::CImage3D() {
	
	m_data = NULL;
	m_width = 0;
	m_height = 0;
	m_depth = 0;
	
	m_scale[0] = 1.0f;
	m_scale[1] = 1.0f;
	m_scale[2] = 1.0f;
}

// ===========
// Destructeur
// ===========
CImage3D::~CImage3D()
{
	if (m_data) delete m_data;
}

// =============================================================
// Fonction lisant un fichier au format RAW contenant les voxels
// soit sur 8bits (t=true) soit sur 16bits (t=false)
// =============================================================
int CImage3D::LoadRAW(char *filename, int w, int h, int d, bool t)
{
	FILE* fp;
	m_width = getNextPowerOfTwo(w);
	m_height = getNextPowerOfTwo(h);
	m_depth = getNextPowerOfTwo(d);
	
	GLushort us;
	GLubyte ub;
	int size = m_width*m_height*m_depth;
	int size2 = w*h*d;
	m_data = new GLubyte[size];

	if (!(fp=fopen(filename,"rb"))){
	printf("Error: unable to open volume\n");
	exit(0);
	}

	printf("Loading volume (%s) : w=%d/%d h=%d/%d d=%d/%d\n", filename, w, m_width, h, m_height, d, m_depth);

	memset(m_data, 0, size);
	
	if (t)
	for(int z=0;z<m_depth;z++)
		for(int y=0;y<m_height;y++)
			for(int x=0;x<m_width;x++)
			{
				//printf("x=%d y=%d z=%d\n",x,y,z);
				if (x>=w || y>=h || z>=d)  m_data[VOXEL(x,y,z)] = 0;
				else {
				fread(&ub, sizeof(GLubyte), 1, fp);
				m_data[VOXEL(x,y,z)] = ub;
				}


			}
	else
	for(int z=0;z<m_depth;z++)
		for(int y=0;y<m_height;y++)
			for(int x=0;x<m_width;x++)
			{
				if (x>=w || y>=h || z>=d)  m_data[VOXEL(x,y,z)] = 0;
				else {
				fread(&us, sizeof(GLushort), 1, fp);
				us=((us&0xff00)>>8)|((us&0xff)<<8);
				m_data[VOXEL(x,y,z)] = (GLubyte)us;
				}
			}

	int m = m_height;
	if (m_width > m) m=m_width;
	else if (m_depth > m) m=m_depth;

	m_scale[0] = ((float)m/(float)m_width);
	m_scale[1] = ((float)m/(float)m_height);
	m_scale[2] = ((float)m/(float)m_depth);

	return 1;
}

// ==================================================
// Retourne le nombre de pixels en largeur de l'image
// ==================================================
short int CImage3D::GetWidth()
{
	return m_width;
}


// ==================================================
// Retourne le nombre de pixels en hauteur de l'image
// ==================================================
short int CImage3D::GetHeight()
{
	return m_height;
}


// =====================================================
// Retourne le nombre de pixels en profondeur de l'image
// =====================================================
short int CImage3D::GetDepth()
{
	return m_depth;
}


// =========================================================
// Creation d'une texture Alpha a partir de l'image courante
// =========================================================
int CImage3D::CreateTexture_Alpha(GLuint textureArray[], GLuint textureID)
{
	int size = m_width*m_height*m_depth;
	GLubyte * tex = new GLubyte[size];

	printf("Loading texture alpha\n");
	
	for(int z=0;z<m_depth;z++)
	{
		for(int y=0;y<m_height;y++)
		{
			for(int x=0;x<m_width;x++)
			{
				if(x==0||x==m_width-1||y==0||y==m_height-1||z==0||z==m_depth-1)
				{
					tex[VOXEL(x,y,z)]=0;
				}	
				else {
					tex[VOXEL(x,y,z)] = (GLubyte)m_data[VOXEL(x,y,z)];
					
				}
			}
		}
	}

	glGenTextures(1, &textureArray[textureID]);

    //load full-size monochrome texture
    glBindTexture(GL_TEXTURE_3D_EXT,textureArray[textureID]);
    glTexParameteri(GL_TEXTURE_3D_EXT,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D_EXT,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D_EXT,GL_TEXTURE_WRAP_R_EXT,GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D_EXT,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_3D_EXT,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
    glTexImage3DEXT(GL_TEXTURE_3D_EXT,0,GL_ALPHA8,m_width,m_height,m_depth,0,GL_ALPHA,GL_UNSIGNED_BYTE,tex);
	
	delete tex;
    
	return 1;
}


// =========================================================
// Creation d'une texture Alpha a partir de l'image courante
// =========================================================
int CImage3D::CreateTexture_TF(GLuint textureArray[], GLuint textureID, GLubyte *tf)
{
	int size = m_width*m_height*m_depth*4;
	GLubyte * tex = new GLubyte[size];

	printf("Loading transfer function texture\n");
	
	for(int z=0;z<m_depth;z++)
	{
		for(int y=0;y<m_height;y++)
		{
			for(int x=0;x<m_width;x++)
			{
				if(x==0||x==m_width-1||y==0||y==m_height-1||z==0||z==m_depth-1)
				{
					tex[4*VOXEL(x,y,z)+0] = 0;
					tex[4*VOXEL(x,y,z)+1] = 0;
					tex[4*VOXEL(x,y,z)+2] = 0;
					tex[4*VOXEL(x,y,z)+3] = 0;
				}	
				else {
					tex[4*VOXEL(x,y,z)+0] = tf[(GLubyte)m_data[VOXEL(x,y,z)]*4];
					tex[4*VOXEL(x,y,z)+1] = tf[(GLubyte)m_data[VOXEL(x,y,z)]*4+1];
					tex[4*VOXEL(x,y,z)+2] = tf[(GLubyte)m_data[VOXEL(x,y,z)]*4+2];
					tex[4*VOXEL(x,y,z)+3] = tf[(GLubyte)m_data[VOXEL(x,y,z)]*4+3];
				}
			}
		}
	}

	if (glIsTexture(textureArray[textureID])) glDeleteTextures(1, &textureArray[textureID]);
	glGenTextures(1, &textureArray[textureID]);

    //load full-size monochrome texture
    glBindTexture(GL_TEXTURE_3D_EXT,textureArray[textureID]);
    glTexParameteri(GL_TEXTURE_3D_EXT,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D_EXT,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D_EXT,GL_TEXTURE_WRAP_R_EXT,GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D_EXT,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_3D_EXT,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexImage3DEXT(GL_TEXTURE_3D_EXT,0,GL_RGBA8,m_width,m_height,m_depth,0,GL_RGBA,GL_UNSIGNED_BYTE,tex);
	
	delete tex;
    
	return 1;
}

// =========================================================
// Creation d'une texture Alpha a partir de l'image courante
// =========================================================
int CImage3D::CreateTexture_TF_PalettedTexture(GLuint textureArray[], GLuint textureID, GLubyte *tf)
{
	int size = m_width*m_height*m_depth;
	GLubyte * tex = new GLubyte[size];

	printf("Loading transfer function texture\n");
	
	for(int z=0;z<m_depth;z++)
	{
		for(int y=0;y<m_height;y++)
		{
			for(int x=0;x<m_width;x++)
			{
				if(x==0||x==m_width-1||y==0||y==m_height-1||z==0||z==m_depth-1)
				{
					tex[VOXEL(x,y,z)]=0;
				}	
				else {
					tex[VOXEL(x,y,z)] = (GLubyte)m_data[VOXEL(x,y,z)];
				}
			}
		}
	}

	if (glIsTexture(textureArray[textureID])) glDeleteTextures(1, &textureArray[textureID]);
	glGenTextures(1, &textureArray[textureID]);

    //load full-size monochrome texture
    glBindTexture(GL_TEXTURE_3D_EXT,textureArray[textureID]);
	glColorTableEXT(GL_TEXTURE_3D_EXT,GL_RGBA8,256,GL_RGBA,GL_UNSIGNED_BYTE,tf);
    glTexParameteri(GL_TEXTURE_3D_EXT,GL_TEXTURE_WRAP_S,GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D_EXT,GL_TEXTURE_WRAP_T,GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D_EXT,GL_TEXTURE_WRAP_R_EXT,GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_3D_EXT,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
    glTexParameteri(GL_TEXTURE_3D_EXT,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
	glTexImage3DEXT(GL_TEXTURE_3D_EXT,0,GL_COLOR_INDEX8_EXT,m_width,m_height,m_depth,0,GL_COLOR_INDEX,GL_UNSIGNED_BYTE,tex);
	delete tex;
    
	return 1;
}

float* CImage3D::GetScale() {
	return m_scale;
}

GLubyte * CImage3D::GetData() {
	return m_data;
}

 
 
 
 
