/*
 * 29/03/2004
 *
 * image3D.h
 * Permet de charger une image 3D (volume) et de crer des textures 3D
 *
 * Author: Thierry Vouriot (Yri)
 * yeri@fr.st
 * http://www.yeri.fr.st
 * thierry.vouriot@free.fr
 * http://thierry.vouriot.free.fr
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#ifndef _IMAGE3D_H
#define _IMAGE3D_H

#ifdef _WIN32
#include <windows.h>
#endif
#include <GL/gl.h>
#include <GL/glu.h>
#include <cstdio>
#include <cstdlib>
#include <cmath>
#include "../ogl/glextensions.h"
#include "../math/math_utils.h"

#define VOXEL(X,Y,Z) ((Z)*m_width*m_height+(Y)*m_width+(X))

// ###############
// Classe CImage3D
// ###############
class CImage3D
{
public:

	// ============
	// Constructeur
	// ============
	CImage3D();

	// ===========
	// Destructeur
	// ===========
	~CImage3D();
	
	// =============================================================
	// Fonction lisant un fichier au format RAW contenant les voxels
	// soit sur 8bits (t=true) soit sur 16bits (t=false)
	// =============================================================
	int LoadRAW(char *filename, int w, int h, int d, bool t);

	// ==================================================
	// Retourne le nombre de pixels en largeur de l'image
	// ==================================================
	short int GetWidth();

	// ==================================================
	// Retourne le nombre de pixels en hauteur de l'image
	// ==================================================
	short int GetHeight();

	// =====================================================
	// Retourne le nombre de pixels en profondeur de l'image
	// =====================================================
	short int GetDepth();

	// =========================================================
	// Creation d'une texture Alpha a partir de l'image courante
	// =========================================================
	int CreateTexture_Alpha(GLuint textureArray[], GLuint textureID);

	// =============================================================================================
	// Creation d'une texture en appliquant la fonction de transfer PseudoColor sur l'image courante
	// =============================================================================================
	int CreateTexture_TF(GLuint textureArray[], GLuint textureID, GLubyte *tf);

	int CreateTexture_TF_PalettedTexture(GLuint textureArray[], GLuint textureID, GLubyte *tf);
	
	float* GetScale();

	GLubyte * GetData();

private:

	// Nombre de pixels en largeur de l'image
	short int m_width;

	// Nombre de pixels en hauteur de l'image
	short int m_height;

	// 
	short int m_depth;

	// Tableau contenant les pixels(voxels) de l'image
	GLubyte * m_data;

	float m_scale[3];

};

#endif


