/*
 * 05/28/2003
 *
 * initGLUT.cpp
 * Thierry Vouriot (Yri)
 * yeri@fr.st
 * http://www.yeri.fr.st
 * thierry.vouriot@free.fr
 * http://thierry.vouriot.free.fr
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "initGLUT.h"

// Fullscreen
bool g_fullscreen = false;

// ======================================================
// Initialisation de GLUT
// (la fenetre, les fonctions callbacks, parametres GLUT)
// ======================================================
void initGLUT()
{
	glutInitDisplayMode(GLUT_DEPTH | GLUT_DOUBLE | GLUT_RGB);
	
	#ifdef _WIN32
	askFullscreen();
	#endif

	if (!g_fullscreen) {
		glutInitWindowSize(WINDOW_WIDTH, WINDOW_HEIGHT);
		glutInitWindowPosition(WINDOW_POS_X, WINDOW_POS_Y);
		glutCreateWindow(APP_NAME);
	}

	// Active la repetition de touches
	glutSetKeyRepeat(GLUT_KEY_REPEAT_ON);
	glutIgnoreKeyRepeat(0); 

	glutDisplayFunc(glutDisplay);
	glutIdleFunc(glutIdle);
	glutReshapeFunc(glutReshape);
	glutKeyboardFunc(glutKeyboard);
	glutSpecialFunc(glutSpecialKeys);

	glutMouseFunc(glutMouse);
	glutMotionFunc(glutMotion);
	glutPassiveMotionFunc(glutPassiveMotion);

}

// ========================
// Initialisation de OpenGL
// ========================
void initOpenGL(void)
{
	/* Affiche des informations sur OpenGL */
	char *exten;
	exten = (char *) glGetString(GL_VENDOR);
	printf("%s\n",exten);
	exten = (char *) glGetString(GL_RENDERER);
	printf("%s\n",exten);
	exten = (char *) glGetString(GL_VERSION);
	printf("%s\n",exten);

	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

	glDepthFunc (GL_LEQUAL);									// The Type Of Depth Testing (Less Or Equal)
	glEnable (GL_DEPTH_TEST);									// Enable Depth Testing
	glEnable(GL_CULL_FACE);										// permet de masquer les faces arrires	
	glShadeModel (GL_SMOOTH);									// Select Smooth Shading
	glHint (GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);			// Set Perspective Calculations To Most Accurate
	glBlendFunc(GL_SRC_ALPHA,GL_ONE);							// Enable Blending
}

// ===============================
// Redimensionnement de la fentre
// ===============================
void glutReshape(int w, int h)
{
	if (h == 0) h = 1;
	glViewport(0, 0, (GLsizei)w, (GLsizei)h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45.0f,(GLfloat)w/(GLfloat)h, 0.1f, 20.0f);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

// =========================================================
// Permet de passer en mode plein ecran ou en mode fenetre
// Utilise GLUT_GAME_MODE disponible uniquement sous Windows
// =========================================================
#ifdef _WIN32
void askFullscreen(void)
{
	char gameMode[14];
	
	/* Demande  l'utilisateur s'il veut le mode plein cran */
	if (!g_fullscreen)
	g_fullscreen = (MessageBox(NULL, "Click Yes to go to full screen (Recommended)", "Start FullScreen?",
						MB_ICONQUESTION | MB_YESNO) == IDYES);
	
	
	if (g_fullscreen) {	// game mode (fullscreen)
		sprintf(gameMode, "%dx%d:%d", FULLSCREEN_WIDTH, FULLSCREEN_HEIGHT, FULLSCREEN_DEPTH);
		glutGameModeString(gameMode);
		// teste si le game mode est possible
		if (glutGameModeGet(GLUT_GAME_MODE_POSSIBLE))
		{
			glutEnterGameMode();
			//glutSetCursor(GLUT_CURSOR_NONE);
		}
		else {
			g_fullscreen = false;
			printf("Error Fullscreen Mode !\n");
		}
	} // mode fentr
}
#endif

// ====================
// Quitte l'application
// ====================
void terminateApplication(void)
{
	#ifdef _WIN32
	if (g_fullscreen) glutLeaveGameMode();
	#endif
	cout << "\nAuteur: Thierry Vouriot\nhttp://thierry.vouriot.free.fr\n";
	exit(0);
}

