/*
 * 29/03/2004
 *
 * main.cpp
 * Thierry Vouriot (Yri)
 * yeri@fr.st
 * http://www.yeri.fr.st
 * thierry.vouriot@free.fr
 * http://thierry.vouriot.free.fr
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "main.h"

CMouse *g_mouse;
CTimerGLUT *g_timer;
GLuint g_textures[1];
GLfloat viewDistance = -4.4;

int slices = 80, lastslices;
float intensity = 0.20;
float density = 0.1;
float scale[3] = {1.0, 1.0, 1.0};
CImage3D *i3d;
CMyGui *mygui;
GLenum filter = GL_LINEAR, lastfilter;
bool aff = true;

float lastPos[3];
GLfloat viewMatrix[16] = {1.0f, 0.0f, 0.0f, 0.0f,
                            0.0f, 1.0f, 0.0f, 0.0f,
                            0.0f, 0.0f, 1.0f, 0.0f,
                            0.0f, 0.0f, 0.0f, 1.0f};

// ====
// Main
// ====
int main(int argc, char ** argv)
{	
	if (argc != 6 && argc != 1) { printf("Usage: visuvo filename width height depth [8|16]\n"); exit(0); }
	glutInit(&argc, argv);
	initGLUT();
	initOpenGL();
	init(argc, argv);
	glutMainLoop();
	return 0;
}

// ========================
// Initialisations diverses
// ========================
void init(int argc, char ** argv)
{
	// Allocation des variables globales
	g_mouse = new CMouse();
	g_timer = new CTimerGLUT();

	// Extensions OpenGL disponibles
	initExtensions();
	CString missing;
	if (!GL_EXT_texture3D_supported)					missing += "GL_EXT_texture3D\n";
    if (missing.getLength() > 0){
		printf("\nThe following required OpenGL extensions isn't supported by the graphic card and driver you're using:\n\n" + missing +
                   "\nThis application will now exit.\n");
		#ifdef _WIN32
		MessageBox(NULL,"The following required OpenGL extensions isn't supported by the graphic card and driver you're using:\n\n" + missing +
                   "\nThis application will now exit.", "Missing extensions", MB_OK);
		#endif
		exit(-1);
    }

	// Volume  charger avec ses paramtres
	char file[1024];
	int w;
	int h;
	int d;
	bool t;
	if (argc == 1) {
		w = h = d = 64;
		t = true;
		strcpy(file,"data/neghip.raw");
	}
	else {
		w = atoi(argv[2]);
		h = atoi(argv[3]);
		d = atoi(argv[4]);
		t = (8 == atoi(argv[5]));
		strcpy(file,argv[1]);
	}
	
	// Chargement du volume
	i3d = new CImage3D();
	i3d->LoadRAW(file, w, h, d, t);
	
	scale[0] =  i3d->GetScale()[0]/2;
	scale[1] =  i3d->GetScale()[1]/2;
	scale[2] =  i3d->GetScale()[2]/2;	

	// Initialisation de l'interface graphique
	mygui = new CMyGui(0, i3d);
	mygui->SetPosition(10, 10);
	mygui->SetShow(MENU);
	mygui->OnClick(guiMouseClick);
	mygui->OnMotion(guiMouseMotion);
	mygui->size->SetLabel("Size: %dx%dx%d", i3d->GetWidth(), i3d->GetHeight(), i3d->GetDepth());
	mygui->density->SetLabel("Density: %.2f", density);
	mygui->intensity->SetLabel("Intensity: %.2f", intensity);
	mygui->slices->SetLabel("Slices: %d", slices);
	mygui->density->SetEnable(false);
	mygui->dplus->SetEnable(false);
	mygui->dminus->SetEnable(false);
	if (GL_EXT_paletted_texture_supported) mygui->tfapply->SetEnable(false);
	if (!GL_EXT_blend_color_supported) mygui->attenuate->SetEnable(false);
	if (!GL_EXT_blend_minmax_supported) mygui->mip->SetEnable(false);
	
	// Cration de la texture 3D
	if (GL_EXT_paletted_texture_supported) i3d->CreateTexture_TF_PalettedTexture(g_textures, 0, mygui->tf->tf);
	else i3d->CreateTexture_TF(g_textures, 0, mygui->tf->tf);
}

// ================================
// Fonction d'affichage de la scene
// ================================
void glutDisplay(void)
{
	int i;
	float mat[16];
	
	// TexGen planes
	float xPlane[]={ 1.0f, 0.0f, 0.0f, 0.0f };
	float yPlane[]={ 0.0f, 1.0f, 0.0f, 0.0f };
	float zPlane[]={ 0.0f, 0.0f, 1.0f, 0.0f };
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// On se dplace  une certaine distance
	glTranslatef(0.0f,0.0f,viewDistance);
	
	// Rotation du model
	glMultMatrixf((GLfloat *) viewMatrix);

	// Mat contient l'inverse de la matrice MODELVIEW
	glGetFloatv(GL_MODELVIEW_MATRIX,mat);
  	matrixInvert(mat);
	//want rotation only, zero the translation
	mat[12]=0.0f;
	mat[13]=0.0f;
	mat[14]=0.0f;
	
	// Pour la gnration des coordonnes des textures
  	glTexGeni(GL_S,GL_TEXTURE_GEN_MODE,GL_EYE_LINEAR);
  	glTexGeni(GL_T,GL_TEXTURE_GEN_MODE,GL_EYE_LINEAR);
  	glTexGeni(GL_R,GL_TEXTURE_GEN_MODE,GL_EYE_LINEAR);  
	
	// On passe  la matrice de texture
	glMatrixMode(GL_TEXTURE);
	glEnable(GL_TEXTURE_3D_EXT);
	glLoadIdentity();
	
	// Applique les transformations  l'objet
	glTranslatef(0.5f, 0.5f, 0.5f);
  	glScalef(scale[0],scale[1],scale[2]);
	
	// On multiplie la matrice TEXTURE par la matrice inverse MODELVIEW
	glMultMatrixf(mat);
	// On applique la translation inverse pour la distance de vue
  	glTranslatef(0.0f,0.0f,-viewDistance);	
		
	// On retourne  la matrice MODELVIEW
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	
	// Dfinition des plans pour TexGen
  	glTexGenfv(GL_S,GL_EYE_PLANE,xPlane);
  	glTexGenfv(GL_T,GL_EYE_PLANE,yPlane);
  	glTexGenfv(GL_R,GL_EYE_PLANE,zPlane);

	// Active la gnration automatique des coordonnes de textures
 	glEnable(GL_TEXTURE_GEN_S);
  	glEnable(GL_TEXTURE_GEN_T);
  	glEnable(GL_TEXTURE_GEN_R);
	
	// Initialise la couleur des tranches du volume
	glColor4f(1.0,1.0,1.0, intensity);
	glEnable(GL_TEXTURE_3D_EXT);
	glBindTexture(GL_TEXTURE_3D_EXT, g_textures[0]);
	
	// Filtrage des textures
	glTexParameteri(GL_TEXTURE_3D_EXT,GL_TEXTURE_MAG_FILTER,filter);
	glTexParameteri(GL_TEXTURE_3D_EXT,GL_TEXTURE_MIN_FILTER,filter);
	
	// Active le test alpha blending
	glEnable(GL_ALPHA_TEST);

	if (!mygui->usetf->IsSelected()) {
		glAlphaFunc(GL_GREATER, density*intensity);
	}
	else glAlphaFunc(GL_GREATER, 0);

	// Choix de la fonction de blending
	if (GL_EXT_blend_minmax_supported) glBlendEquationEXT(GL_FUNC_ADD_EXT);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	if (mygui->over->IsSelected()) {
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	}
	else if (mygui->attenuate->IsSelected()) {
		glBlendFunc(GL_CONSTANT_ALPHA_EXT, GL_ONE);
		glBlendColorEXT(1.f, 1.f, 1.f, 1.f/slices*3.0f);
	}
	else if (mygui->mip->IsSelected()) {
		glBlendEquationEXT(GL_MAX_EXT);
	}
	glEnable(GL_BLEND);
	
	// On se dplace  une certaine distance
	glTranslatef(0.0f,0.0f,viewDistance);
	
	// Dessine les tranches du volumes
	for (i=0; i < slices; i++)
	{
		glPushMatrix();
		// Z-translation pour dessiner la tranche courante
		glTranslatef(0.0, 0.0, -1.0 + (float)(i)*2.0f/(float)(slices-1));
		
		// Dessin de la tranche
		// Les coordonnes de textures sont assignes automatiquement
		// grace  l'utilisation de TexGen
		glBegin(GL_QUADS);
			glVertex3f(-1.0, -1.0, 0.0);
			glVertex3f(1.0, -1.0, 0.0);
			glVertex3f(1.0, 1.0, 0.0);
			glVertex3f(-1.0, 1.0, 0.0);
		glEnd();
		glPopMatrix();
	}

	glDisable(GL_ALPHA_TEST);
	glDisable(GL_BLEND);
	glDisable(GL_TEXTURE_3D_EXT);
	glPopMatrix();
	
	// Desactive la gnration automatique des coordonnes de textures
 	glDisable(GL_TEXTURE_GEN_S);
  	glDisable(GL_TEXTURE_GEN_T);
  	glDisable(GL_TEXTURE_GEN_R);

	if (aff) {
		mygui->SetVisible(true);
		// Dessine un cube autour de la texture
		glColor4f(0.0f,0.0f,0.7f,1.0f);
		if (mygui->show != NO)
		glutWireCube(2.0);

		// Dessine l'interface utilisateur
		glBlendFunc(GL_SRC_ALPHA, GL_ONE);
		glEnable(GL_BLEND);
	
		mygui->fps->SetLabel("%4d fps", g_timer->GetFPS());
		mygui->Draw(glutGet(GLUT_WINDOW_WIDTH),glutGet(GLUT_WINDOW_HEIGHT));
	}
	else mygui->SetVisible(false);

	glFlush();
	glutSwapBuffers();
}

// ===========================================================
// Fonction excut lorsque le programme n'a plus rien  faire
// utilisee pour les animations
// ===========================================================
void glutIdle(void)
{
	// Calcul le framerate
	g_timer->CalcFPS();
	
	glutPostRedisplay();
}

// ==================================
// Gestion du clavier (touches ascii)
// ==================================
void glutKeyboard(unsigned char key, int x, int y)
{
	switch(key) {
		case ESC:	// touche escape
			terminateApplication();
			break;
		case 'a':
		case 'A':
			intensity += 0.01;
			if (intensity > 1) intensity = 1.0;
			mygui->intensity->SetLabel("Intensity: %.2f", intensity);
			break;
		case 'q':
		case 'Q':
			intensity -= 0.01;
			if (intensity < 0) intensity = 0.0;
			mygui->intensity->SetLabel("Intensity: %.2f", intensity);
			break;
		case 'z':
		case 'Z':
			density += 0.01;
			if (density > 1) density = 1.0;
			mygui->density->SetLabel("Density: %.2f", density);
			break;
		case 's':
		case 'S':
			density -= 0.01;
			if (density < 0) density = 0.0;
			mygui->density->SetLabel("Density: %.2f", density);
			break;
		case 'e':
		case 'E':
			slices += 10;
			if (slices > 200) slices = 200;
			mygui->slices->SetLabel("Slices: %d", slices);
			break;
		case 'd':
		case 'D':
			slices -= 10;
			if (slices < 10) slices = 10;
			mygui->slices->SetLabel("Slices: %d", slices);
			break;
		case ' ':
			aff = !aff;
			break;
	}
}

// ======================================
// Gestion du clavier (touches non ascii)
// ======================================
void glutSpecialKeys(int key, int x, int y)
{
	
}

// =========================================
// Gestion de la souris (clicks des boutons)
// =========================================
void glutMouse(int button, int state, int x, int y)
{
	mygui->Mouse(button, state, x, glutGet(GLUT_WINDOW_HEIGHT)-y);

	if (button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
		g_mouse->left_button = true;
		ptov(x, y, glutGet(GLUT_WINDOW_WIDTH), glutGet(GLUT_WINDOW_HEIGHT), lastPos);
		lastslices = slices;
		lastfilter = filter;
	}
		
	else if (button == GLUT_LEFT_BUTTON && state == GLUT_UP) {
		g_mouse->left_button = false;
		slices = lastslices;
		filter = lastfilter;
	}
		
	else if (button == GLUT_RIGHT_BUTTON && state == GLUT_DOWN) {
		g_mouse->right_button = true;
	}
		
	else if (button == GLUT_RIGHT_BUTTON && state == GLUT_UP)
		g_mouse->right_button = false;
	
	// sauvegarde de la position lors de l'appui sur un bouton
	g_mouse->x = x;
	g_mouse->y = y;

	glutPostRedisplay();
}

// ================================================================
// Gestion des mouvements de la souris lorsqu'un bouton est enfonce
// ================================================================
void glutMotion(int x, int y)
{
	// bouton gauche = rotation
	if (g_mouse->left_button) {
		mygui->MouseMotion(GLUT_LEFT_BUTTON, g_mouse->x, glutGet(GLUT_WINDOW_HEIGHT)-g_mouse->y, x, glutGet(GLUT_WINDOW_HEIGHT)-y);
		if (!mygui->tf->selected) {
			
			float curPos[3], dx, dy, dz;
			float angle=0.0f, axis[3]={1.0f,0.0f,0.0f};
			
			ptov(x, y, glutGet(GLUT_WINDOW_WIDTH), glutGet(GLUT_WINDOW_HEIGHT), curPos);

			dx = curPos[0] - lastPos[0];
			dy = curPos[1] - lastPos[1];
			dz = curPos[2] - lastPos[2];
	
			angle = 90.0 * sqrt(dx*dx + dy*dy + dz*dz);

			axis[0] = lastPos[1]*curPos[2] - lastPos[2]*curPos[1];
			axis[1] = lastPos[2]*curPos[0] - lastPos[0]*curPos[2];
			axis[2] = lastPos[0]*curPos[1] - lastPos[1]*curPos[0];
     
			lastPos[0] = curPos[0];
			lastPos[1] = curPos[1];
			lastPos[2] = curPos[2];
     
			glPushMatrix();
			glLoadIdentity();
			glRotatef(angle, axis[0], axis[1], axis[2]);
			glMultMatrixf((GLfloat *) viewMatrix);
			glGetFloatv(GL_MODELVIEW_MATRIX, (GLfloat *) viewMatrix);
			glPopMatrix();

#ifdef OPTIMIZE_MODE
			slices = (int)(lastslices / 1.5f);
			if (slices < 10) slices = 10;
			filter = GL_NEAREST;
#endif
		}

		g_mouse->x = x;
		g_mouse->y = y;
		glutPostRedisplay();
	}
	// bouton droit = zoom
	if (g_mouse->right_button) {
		mygui->MouseMotion(GLUT_RIGHT_BUTTON, g_mouse->x, glutGet(GLUT_WINDOW_HEIGHT)-g_mouse->y, x, glutGet(GLUT_WINDOW_HEIGHT)-y);

		viewDistance += ((float) (y - g_mouse->y)) / 20.0;
		if (viewDistance > -1.0)
			viewDistance = -1.0;
		g_mouse->y = y;
		glutPostRedisplay();
	}
}

// ===================================
// Gestion des mouvements de la souris
// ===================================
void glutPassiveMotion(int x, int y)
{
	mygui->MousePassiveMotion(x, glutGet(GLUT_WINDOW_HEIGHT)-y);

	g_mouse->x = x;
	g_mouse->y = y;
}

// =======================================================
// Gestion des clicks de souris sur les lments de la GUI
// =======================================================
void guiMouseClick(int id) {

	switch (id) {

		case BUTTON_EXIT:
			terminateApplication();
			break;
		case CHECKBOX_FILTER:
			if (filter == GL_LINEAR) filter = GL_NEAREST;
			else filter = GL_LINEAR;
			break;
		case BUTTON_TFAPPLY:
			if (mygui->usetf->IsSelected()) {
				mygui->tf->Maj();
				i3d->CreateTexture_TF(g_textures, 0, mygui->tf->tf);
			}
			break;
		case BUTTON_TFINIT:
			mygui->tf->Init(i3d);
			break;
		case BUTTON_MENU:
			if (mygui->show == 1) mygui->SetShow(NO); else mygui->SetShow(MENU);
			break;
		case BUTTON_INFOS:
			if (mygui->show == 2) mygui->SetShow(NO); else mygui->SetShow(INFOS);
			break;
		case CHECKBOX_USETF:
			if (mygui->usetf->IsSelected()) {
				mygui->tf->Maj();
				i3d->CreateTexture_TF(g_textures, 0, mygui->tf->tf);
				mygui->density->SetEnable(false);
				mygui->dplus->SetEnable(false);
				mygui->dminus->SetEnable(false);
			}
			else {
				i3d->CreateTexture_Alpha(g_textures, 0);
				mygui->density->SetEnable(true);
				mygui->dplus->SetEnable(true);
				mygui->dminus->SetEnable(true);
			}
			break;
		case BUTTON_DPLUS:
			density += 0.05;
			if (density > 1) density = 1.0;
			mygui->density->SetLabel("Density: %.2f", density);
			break;
		case BUTTON_DMINUS:
			density -= 0.05;
			if (density < 0) density = 0.0;
			mygui->density->SetLabel("Density: %.2f", density);
			break;
		case BUTTON_SPLUS:
			slices += 10;
			if (slices > 200) slices = 200;
			mygui->slices->SetLabel("Slices: %d", slices);
			break;
		case BUTTON_SMINUS:
			slices -= 10;
			if (slices < 10) slices = 10;
			mygui->slices->SetLabel("Slices: %d", slices);
			break;
		case BUTTON_IPLUS:
			intensity += 0.05;
			if (intensity > 1) intensity = 1.0;
			mygui->intensity->SetLabel("Intensity: %.2f", intensity);
			break;
		case BUTTON_IMINUS:
			intensity -= 0.05;
			if (intensity < 0) intensity = 0.0;
			mygui->intensity->SetLabel("Intensity: %.2f", intensity);
			break;
	}
}

// ================================================================
// Gestion des mouvements de la souris lorsqu'un bouton est enfonce
// sur les lments de la GUI
// ================================================================
void guiMouseMotion(int id, int x, int y) {

	switch (id) {
		case TRANSFERFUNCTION:
			/*if (mygui->usetf->IsSelected()) {
				mygui->tf->Maj();
				i3d->CreateTexture_TF(g_textures, 0, mygui->tf->tf);
			}*/
			if (GL_EXT_paletted_texture_supported) {
				mygui->tf->Maj();
				glBindTexture(GL_TEXTURE_3D_EXT, g_textures[0]);
				glColorTableEXT(GL_TEXTURE_3D_EXT,GL_RGBA8,256,GL_RGBA,GL_UNSIGNED_BYTE,mygui->tf->tf);
			}
			break;
	}
}

